use std::str::FromStr;

use chrono::{DateTime, NaiveDateTime, TimeZone, Utc};
use once_cell::sync::Lazy;
use reqwest::header::{HeaderMap, HeaderValue};
use reqwest::Client;
use warp::reject::Reject;
use warp::Rejection;

static CLIENT: Lazy<Client> = Lazy::new(|| {
    let mut headers = HeaderMap::with_capacity(1);

    // teletrader refuses to accept connections without the user agent
    headers.append("user-agent", HeaderValue::from_str("Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/90.0.4430.85 Safari/537.36").unwrap());

    Client::builder().default_headers(headers).build().unwrap()
});

#[derive(Serialize)]
struct TTSResponse {
    copyright: &'static str,
    data: Vec<Data>,
}

#[derive(Serialize)]
struct Data {
    date: DateTime<Utc>,
    price: f64,
    high: f64,
    low: f64,
    volume: u64,
}

#[derive(Debug)]
struct MsgRejection {
    msg: String,
}

impl Reject for MsgRejection {}

impl MsgRejection {
    pub fn reject(msg: String) -> Rejection {
        warp::reject::custom(MsgRejection { msg })
    }
}

pub async fn proxy_tts(tts: String) -> Result<Box<dyn warp::Reply>, warp::Rejection> {
    proxy(format!("tts-{}", tts)).await
}

pub async fn proxy(id: String) -> Result<Box<dyn warp::Reply>, warp::Rejection> {
    info!("serving request for ID {}", id);

    let data = fetch_data(&id)
        .await
        .map_err(|err| MsgRejection::reject(format!("error during data fetching: {:?}", err)))?;
    let response = deserialize_data(&data).map_err(|err| {
        MsgRejection::reject(format!("error during data deserialization: {:?}", err))
    })?;

    Ok(Box::new(warp::reply::json(&response)))
}

async fn fetch_data(id: &str) -> anyhow::Result<String> {
    // todo: make time range configurable?
    
    let url = format!("https://www.teletrader.com/ChartDataHistory.aspx?request=HISTORY%20{id}%20DAILY%201000%20100000%209900000%209900000&dataLoader=ttws", id = id);
    let response = CLIENT.get(url).send().await?;
    response.text().await.map_err(|err| err.into())
}

fn deserialize_data(serialized: &str) -> anyhow::Result<TTSResponse> {
    let data = serialized
        .lines()
        .skip(1) // skip BEGINDATA
        .filter(|&line| line != "---packet end---") // skip ---packet end---
        .map(|data| data.split(';'))
        .map(|mut array| {
            let mut next = || array.next().ok_or_else(|| anyhow::Error::msg("not found"));

            // date;open;high;low;price;volume
            Ok((
                next()?, // date
                next()?, // open
                next()?, // high
                next()?, // low
                next()?, // price
                next()?, // volume
            ))
        })
        .map(|result: anyhow::Result<_>| {
            let (date, open, high, low, price, volume) = result?;

            Ok((
                Utc.from_local_datetime(&NaiveDateTime::parse_from_str(date, "%Y-%m-%d-%H-%M")?)
                    .single()
                    .ok_or_else(|| anyhow::Error::msg("not a single datetime"))?,
                f64::from_str(open)?,
                f64::from_str(high)?,
                f64::from_str(low)?,
                f64::from_str(price)?,
                u64::from_str(volume)?,
            ))
        })
        .map(|result: anyhow::Result<_>| {
            let (date, _open, high, low, price, volume) = result?;

            Ok(Data {
                date,
                price,
                high,
                low,
                volume,
            })
        })
        .collect::<anyhow::Result<_>>()?;

    Ok(TTSResponse {
        copyright: "This data is property of baha or the respective licensor. The terms of use can be accessed at https://www.teletrader.com/teletrader-products/terms_black",
        data,
    })
}
